@@ -6,7 +6,7 @@ use Try::Tiny;
 require Scalar::Util;
 require Symbol;
 
-our $VERSION= '0.03';
+our $VERSION= '0.04';
 our @CARP_NOT= qw( DataStore::CAS::File DataStore::CAS::VirtualHandle );
 
 # ABSTRACT: Abstract base class for Content Addressable Storage
@@ -21,28 +21,28 @@ sub _build_hash_of_null {
 }
 
 
-sub calculate_hash {
-	my $self= shift;
-	Digest->new($self->digest)->add(ref $_[0]? ${$_[0]} : $_[0])->hexdigest;
-}
-
-sub calculate_file_hash {
-	my ($self, $file)= @_;
-	Digest->new($self->digest)->addfile($file)->hexdigest;
-}
-
-
 requires 'get';
 
 
+sub _thing_stringifies_to_filename {
+	my $ref= ref $_[0];
+	(!$ref && defined $_[0])
+		|| $ref->isa('Path::Class::File')
+		|| $ref->isa('Path::Tiny')
+		|| $ref->isa('File::Temp')
+}
+
 sub put {
+	my $ref= ref $_[1];
 	goto $_[0]->can('put_scalar')
-		if !ref $_[1] || ref $_[1] eq 'SCALAR';
+		if !$ref || $ref eq 'SCALAR';
 	goto $_[0]->can('put_file')
-		if ref($_[1])->isa('DataStore::CAS::File') or ref($_[1])->isa('Path::Class::File');
+		if $ref->isa('DataStore::CAS::File')
+		or _thing_stringifies_to_filename($_[1]);
 	goto $_[0]->can('put_handle')
-		if ref($_[1])->isa('IO::Handle') or (Scalar::Util::reftype($_[1]) eq 'GLOB');
-	croak("Can't 'put' object of type ".ref($_[1]));
+		if $ref->isa('IO::Handle')
+		or Scalar::Util::reftype($_[1]) eq 'GLOB';
+	croak("Can't 'put' object of type $ref");
 }
 
 
@@ -80,7 +80,12 @@ sub put_scalar {
 
 sub put_file {
 	my ($self, $file, $flags)= @_;
-	my $is_cas_file= ref $file && ref($file)->isa('DataStore::CAS::File');
+	my $ref= ref $file;
+	my $is_cas_file= $ref && $ref->isa('DataStore::CAS::File');
+	my $is_filename= _thing_stringifies_to_filename($file);
+	croak "Unhandled argument to put_file: ".($file||'(undef)')
+		unless defined $file && ($is_cas_file || $is_filename);
+
 	my %known_hashes= $flags->{known_hashes}? %{$flags->{known_hashes}} : ();
 	# Apply reuse_hash feature, if requested
 	if ($is_cas_file && $flags->{reuse_hash}) {
@@ -90,13 +95,10 @@ sub put_file {
 	# It is probably better to read a file twice than to write one that
 	# doesn't need to be written.
 	# ...but can't do better than ->put_handle unless the file is a real file.
-	my $fname= !ref $file? $file
-		: $is_cas_file? (
-			$file->can('local_file') and length $file->local_file? $file->local_file : undef
-		)
-		: index(ref($file), 'Path') >= 0? "$file"
+	my $fname= $is_filename? "$file"
+		: $is_cas_file && $file->can('local_file')? $file->local_file
 		: undef;
-	if ($fname && -f $fname) {
+	if ($known_hashes{$self->digest} || (defined $fname && -f $fname)) {
 		# Calculate the hash if it wasn't given.
 		my $hash= ($known_hashes{$self->digest} ||= $self->calculate_file_hash($fname));
 		# Avoid unnecessary work
@@ -104,7 +106,7 @@ sub put_file {
 			$flags->{stats}{dup_file_count}++
 				if $flags->{stats};
 			$self->_unlink_source_file($file, $flags)
-				if $flags->{move};
+				if $flags->{move} && defined $fname;
 			return $hash;
 		}
 		# Save hash for next step
@@ -114,11 +116,14 @@ sub put_file {
 	if ($is_cas_file) {
 		$fh= $file->open or croak "Can't open '$file': $!";
 	}
-	elsif (ref($file) && ref($file)->can('open')) {
-		$fh= $file->open('r') or croak "Can't open '$file': $!";
+	elsif ($ref && $ref->can('openr')) {
+		$fh= $file->openr or croak "Can't open '$file': $!";
+	}
+	elsif ($is_filename) {
+		open($fh, '<', $fname) or croak "Can't open '$fname': $!";
 	}
 	else {
-		open($fh, '<', "$fname") or croak "Can't open '$fname': $!";
+		croak "Don't know how to open '$file'";
 	}
 	my $hash= $self->put_handle($fh, $flags);
 	$self->_unlink_source_file($file, $flags)
@@ -135,6 +140,11 @@ sub _unlink_source_file {
 			."If you really want to do this, pass \$file->local_name and then"
 			." delete the cas entry yourself.";
 	} else {
+		if (ref $file && ref($file)->isa('File::Temp')) {
+			# The Simple backend closes File::Temp files to ensure they don't
+			# get written to any more. so match that behavior here.
+			$file->close;
+		}
 		unlink "$file" or croak "unlink($file): $!"
 	}
 }
@@ -171,6 +181,19 @@ sub new_write_handle {
 requires 'commit_write_handle';
 
 
+sub calculate_hash {
+	my $self= shift;
+	Digest->new($self->digest)->add(ref $_[0]? ${$_[0]} : $_[0])->hexdigest;
+}
+
+sub calculate_file_hash {
+	my ($self, $file)= @_;
+	open my $fh, '<', $file or croak "open($file): $!";
+	binmode $fh;
+	Digest->new($self->digest)->addfile($fh)->hexdigest;
+}
+
+
 sub validate {
 	my ($self, $hash, $flags)= @_;
 
@@ -211,7 +234,7 @@ package DataStore::CAS::File;
 use strict;
 use warnings;
 
-our $VERSION= '0.03';
+our $VERSION= '0.04';
 
 sub store { $_[0]{store} }
 sub hash  { $_[0]{hash} }
@@ -247,7 +270,7 @@ package DataStore::CAS::VirtualHandle;
 use strict;
 use warnings;
 
-our $VERSION= '0.03';
+our $VERSION= '0.04';
 
 sub new {
 	my ($class, $cas, $fields)= @_;
@@ -350,7 +373,7 @@ use strict;
 use warnings;
 use parent -norequire => 'DataStore::CAS::VirtualHandle';
 
-our $VERSION= '0.03';
+our $VERSION= '0.04';
 
 # For write-handles, commit data to the CAS and return the digest hash for it.
 sub commit   { $_[0]->_cas->commit_write_handle(@_) }
@@ -367,6 +390,7 @@ sub eof      { return 1; }
 sub read     { return 0; }
 sub readline { return undef; }
 
+
 1;
 
 __END__
@@ -381,7 +405,7 @@ DataStore::CAS - Abstract base class for Content Addressable Storage
 
 =head1 VERSION
 
-version 0.03
+version 0.04
 
 =head1 SYNOPSIS
 
@@ -393,13 +417,22 @@ version 0.03
   );
   
   # Store content, and get its hash code
-  my $hash= $cas->put_scalar("Blah");
+  my $hash0= $cas->put($something_ambiguous);
+  my $hash1= $cas->put_scalar($data_bytes);
+  my $hash2= $cas->put_file($filename);
+  my $hash3= $cas->put_handle(\*STDIN);
   
-  # Retrieve a reference to that content
+  my $writer= $cas->new_write_handle;
+  for (1..10) {
+    $writer->print($_);
+  }
+  my $hash4= $writer->commit;
+  
+  # Retrieve a reference to that content, or undef for unknown hash
   my $file= $cas->get($hash);
   
   # Inspect the file's attributes
-  $file->size < 1024*1024 or die "Use a smaller file";
+  say "File is " . $file->size . " bytes";
   
   # Open a handle to that file (possibly returning a virtual file handle)
   my $handle= $file->open;
@@ -416,17 +449,17 @@ With a good message digest algorithm, one checksum will statistically only
 ever refer to one file, even though the permutations of the checksum are
 tiny compared to all the permutations of bytes that they can represent.
 
-Perl uses the term 'hash' to refer to a mapping of key/value pairs, which
-creates a little confusion.  The documentation of this and related modules
-try to use the phrase "digest hash" to clarify when we are referring to the
-output of a digest function vs. a perl key-value mapping.
-
 In short, a CAS is a key/value mapping where small-ish keys are determined
 from large-ish data but no two pieces of data will ever end up with the same
 key, thanks to astronomical probabilities.  You can then use the small-ish
 key as a reference to the large chunk of data, as a sort of compression
 technique.
 
+Note: Perl uses the term 'hash' to refer to a mapping of key/value pairs,
+which creates a little confusion.  The documentation of this and related
+modules try to use the phrase "digest hash" to clarify when we are referring
+to the output of a digest function vs. a perl key-value hash table.
+
 =head1 PURPOSE
 
 One great use for CAS is finding and merging duplicated content.  If you
@@ -453,14 +486,6 @@ subclass will need to provide a few additional methods like L</calculate_hash>.
 
 The digest hash of the empty string.
 
-=head2 calculate_hash
-
-Return the hash of a scalar (or scalar ref) in memory.
-
-=head2 calculate_file_hash
-
-Return the hash of a file on disk.
-
 =head1 METHODS
 
 =head2 get
@@ -488,7 +513,8 @@ Scalars and Scalar-refs are passed to L</put_scalar>.
 
 =item *
 
-Instances of L<DataStore::CAS::File> or L<Path::Class::File> are passed to L</put_file>.
+Instances of L<DataStore::CAS::File>, L<Path::Class::File>, L<Path::Tiny>,
+or L<Fie::Temp> are passed to L</put_file>.
 
 =item *
 
@@ -583,7 +609,9 @@ Additional flags:
 If move is true, and the CAS is backed by plain files on the same filesystem,
 it will move the file into the CAS, possibly changing its owner and permissions.
 Even if the file can't be moved, C<put_file> will attempt to unlink it, and die
-on failure.
+on failure.  Note: If you use this option with a L<File::Temp> object, this closes
+the file handle to ensure that no further writes or fd-operations are applied
+to the file which is now part of your read-only CAS.
 
 =item hardlink => $bool
 
@@ -668,6 +696,14 @@ This closes the given write-handle, and then finishes calculating its digest
 hash, and then stores it into the CAS (unless the handle was created with the
 dry_run flag).  It returns the digest_hash of the data.
 
+=head2 calculate_hash
+
+Return the hash of a scalar (or scalar ref) in memory.
+
+=head2 calculate_file_hash
+
+Return the hash of a file on disk.
+
 =head2 validate
 
   $bool_valid= $cas->validate( $digest_hash, \%optional_flags )
@@ -782,6 +818,12 @@ filehandle.  Note that returned handles are 'raw' by default.
 
 =back
 
+=head1 THANKS
+
+Portions of this software were funded by
+L<Clippard Instrument Laboratory|https://www.clippard.com>.
+Thanks for supporting Open Source.
+
 =head1 AUTHOR
 
 Michael Conrad <mconrad@intellitree.com>
