@@ -8,7 +8,7 @@ use File::Spec 3.33;
 use File::Spec::Functions 'catfile', 'catdir', 'canonpath';
 use File::Temp 0.22 ();
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 our @CARP_NOT= qw( DataStore::CAS DataStore::CAS::File DataStore::CAS::VirtualHandle );
 
 # ABSTRACT: Simple file/directory based CAS implementation
@@ -116,10 +116,24 @@ sub create_store {
 
 	my $conf_dir= catdir($params{path}, 'conf');
 	mkdir($conf_dir) or croak "mkdir($conf_dir): $!";
-	$class->_write_config_setting($params{path}, 'VERSION', $class.' '.$VERSION."\n");
+	$class->_write_config_setting($params{path}, 'VERSION', $class->_hierarchy_version);
 	$class->_write_config_setting($params{path}, 'digest', $params{digest}."\n");
 	$class->_write_config_setting($params{path}, 'fanout', join(' ', @{$params{fanout}})."\n");
 }
+sub _hierarchy_version {
+	my $class= ref $_[0] || $_[0];
+	my $out= '';
+	# record the version of any class hierarchy which "isa DataStore::CAS::Simple"
+	my $hier= mro::get_linear_isa($class);
+	for (grep $_->isa(__PACKAGE__), @$hier) {
+		if (!$_->VERSION) {
+			warn "Package '$_' lacks a VERSION, weakening the protection of DataStore::CAS::Simple's versioned storage directory.";
+		} else {
+			$out .= $_ . ' ' . $_->VERSION . "\n";
+		}
+	}
+	return $out;
+}
 
 # This method loads the digest and fanout configuration and validates it
 # It is called during the constructor.
@@ -303,14 +317,16 @@ sub get {
 
 sub put_file {
 	my ($self, $file, $flags)= @_;
-	my $is_cas_file= ref $file && ref($file)->isa('DataStore::CAS::File');
+	my $ref= ref $file;
+	my $is_cas_file= $ref && $ref->isa('DataStore::CAS::File');
+	my $is_filename= DataStore::CAS::_thing_stringifies_to_filename($file);
+	croak "Unhandled argument to put_file: ".($file||'(undef)')
+		unless defined $file && ($is_cas_file || $is_filename);
 	
 	# Can only optimize if source is a real file
 	if ($flags->{hardlink} || ($flags->{move} && !$is_cas_file)) {
-		my $fname= 
-			($is_cas_file && $file->can('local_file') and length $file->local_file)? $file->local_file
-			: (ref $file && ref($file)->isa('Path::Class::File'))? "$file"
-			: (!ref $file)? $file
+		my $fname= $is_filename? "$file"
+			: $is_cas_file && $file->can('local_file')? $file->local_file
 			: undef;
 		if ($fname && -f $fname) {
 			my %known_hashes= $flags->{known_hashes}? %{$flags->{known_hashes}} : ();
@@ -348,6 +364,11 @@ sub _try_put_move {
 	my $dest= $self->path_for_hash($hash,1);
 	my $tmp= "$dest.tmp";
 	return 0 unless rename($file, $tmp);
+	if (ref $file && ref($file)->isa('File::Temp')) {
+		# File::Temp creates a writable handle, and operates on the
+		# file using 'fd___' functions, so it needs closed to be safe.
+		$file->close;
+	}
 	# Need to be able to change ownership to current user and remove write bits.
 	try {
 		my ($mode, $uid, $gid)= (stat $tmp)[2,4,5]
@@ -619,7 +640,7 @@ DataStore::CAS::Simple - Simple file/directory based CAS implementation
 
 =head1 VERSION
 
-version 0.03
+version 0.04
 
 =head1 DESCRIPTION
 
